"""
Action Functions
================

This module defines the functions that are used to perform actions on the
environment. The functions are used in the `EnvConfig` class to specify the
actions that can be performed on the environment.
"""
