"""
eprllib
=======

The package `eprllib` provides a set of tools for the development of energy performance
research in the context of the Python programming language. The package is structured
as follows:

.. toctree::
   :maxdepth: 1
   :caption: Contents:
   
   ActionFunctions
   Agents
   Env
   EpisodeFunctions
   ExampleFiles
   PostProcess
   PreProcess
   RewardFunctions
   Tools

"""